/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
 * autor
 * Michal Podsiadlik
 * michal at gov.one.pl
 */
#include <qpopupmenu.h>
#include <qtooltip.h>
#include <qcheckbox.h>
#include <qspinbox.h>
#include <qtooltip.h>

#include <gadu.h>
#include <debug.h>
#include <config_dialog.h>
#include <userbox.h>
#include <kadu_text_browser.h>
#include <icons_manager.h>
#include <action.h>

#include "tabs.h"

extern "C" int tabs_init()
{
	tabs_manager=new TabsManager();
	return 0;
}

extern "C" void tabs_close()
{
	delete tabs_manager;
}

TabsManager::TabsManager() : QObject()
{
	kdebugf();

	connect(chat_manager, SIGNAL(chatCreated(const UserGroup*)),
			this, SLOT(onNewChat(const UserGroup*)));
	connect(chat_manager, SIGNAL(chatDestroying(const UserGroup*)),
			this, SLOT(onDestroyingChat(const UserGroup*)));

	connect(chat_manager, SIGNAL(chatOpen(UserListElements)),
			this, SLOT(onOpenChat(UserListElements)));

	connect(userlist, SIGNAL(statusChanged(UserListElement, QString, const UserStatus &, bool, bool)),
			this, SLOT(onStatusChanged(UserListElement, QString, const UserStatus &, bool, bool)));
	connect(gadu, SIGNAL(chatMsgReceived0(Protocol*, UserListElements, const QString&, time_t, bool&)),
			this, SLOT(onChatMsgReceived(Protocol*, UserListElements, const QString&, time_t, bool&)));

	connect(&timer, SIGNAL(timeout()),
			this, SLOT(onTimer()));

	// przeniesienie starej konfiguracji skrotow z Chat do ShortCuts
	config_file.addVariable("ShortCuts", "MoveTabLeft", config_file.readEntry("Chat", "MoveTabLeft"));
	config_file.addVariable("ShortCuts", "MoveTabRight", config_file.readEntry("Chat", "MoveTabRight"));
	config_file.addVariable("ShortCuts", "SwitchTabLeft", config_file.readEntry("Chat", "SwitchTabLeft"));
	config_file.addVariable("ShortCuts", "SwitchTabRight", config_file.readEntry("Chat", "SwitchTabRight"));

	ConfigDialog::addVGroupBox("Chat", "Chat",
			QT_TRANSLATE_NOOP("@default", "Tabs"));
	ConfigDialog::addCheckBox("Chat", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Use tabs by default"), "DefaultTabs");
	ConfigDialog::addSpinBox("Chat", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Minimum number of tabs"), "MinTabs", 1, 255, 1, 2);
	ConfigDialog::addCheckBox("Chat", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Tabs below chats"), "TabsBelowChats");
	ConfigDialog::addCheckBox("Chat", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Auto tab change"), "AutoTabChange");
	ConfigDialog::addCheckBox("Chat", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Conferences in tabs"), "ConferencesInTabs", true);
	ConfigDialog::addVGroupBox("ShortCuts", "ShortCuts",
			QT_TRANSLATE_NOOP("@default", "Tabs"));
	ConfigDialog::addHotKeyEdit("ShortCuts", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Move tab left"), "MoveTabLeft", "Ctrl+Shift+Left");
	ConfigDialog::addHotKeyEdit("ShortCuts", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Move tab right"), "MoveTabRight", "Ctrl+Shift+Right");
	ConfigDialog::addHotKeyEdit("ShortCuts", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Switch to prev. tab"), "SwitchTabLeft", "Shift+Left");
	ConfigDialog::addHotKeyEdit("ShortCuts", "Tabs",
			QT_TRANSLATE_NOOP("@default", "Switch to next tab"), "SwitchTabRight", "Shift+Right");

	ConfigDialog::registerSlotOnCreateTab("Chat", this,SLOT(onCreateConfigDialog()));
	ConfigDialog::registerSlotOnApplyTab("Chat", this, SLOT(onApplyConfig()));

	UserBox::userboxmenu->addItemAtPos(1, "OpenChat",
			tr("Open in new tab"), this, SLOT(onNewTab()));
	menuitem=UserBox::userboxmenu->getItem(tr("Open in new tab"));

	action = new Action(
		icons_manager->loadIcon(dataPath("kadu/modules/data/tabs/attach.png")),
		tr("Attach chat to tabs"), "tab_attach_action", Action::TypeChat);

	action->setToggleAction(true);

	connect(action, SIGNAL(activated(const UserGroup*, const QWidget*, bool)),
			this, SLOT(onTabAttach(const UserGroup*, const QWidget*, bool)));
	connect(action, SIGNAL(addedToToolbar(const UserGroup*, ToolButton*, ToolBar*)),
			this, SLOT(onAddedToToolbar(const UserGroup*)));


	KaduActions.insert("tab_attach_action", action);

	tabdialog=new TabWidget();
	connect(tabdialog, SIGNAL(currentChanged(QWidget *)),
			this, SLOT(onTabChange(QWidget *)));
	connect(tabdialog, SIGNAL(contextMenu(QWidget*,const QPoint&)),
			this, SLOT(onContextMenu(QWidget*,const QPoint&)));
	connect(tabdialog, SIGNAL(deleteTab(QWidget*)),
			this, SLOT(onDeleteTab(QWidget*)));

	loadGeometry(tabdialog, "Chat", "TabWindowsGeometry", 30, 30, 400, 400);

	connect(UserBox::userboxmenu, SIGNAL(popup()), this, SLOT(onPopupMenu()));

	//aktywacja minimalnej liczby kart
	ConfigDialog::connectSlot("Chat", "Use tabs by default", SIGNAL(toggled(bool)), this, SLOT(onToggledDefaultTabs(bool)));

	// pozycja tabw
	onApplyConfig();

	// zrb mi menu :>
	makePopupMenu();

	no_tabs=false;
	autoswith=false;

	// sprawdzanie czy przed zaadowaniem moduu zostay utworzone jakie chaty i dodanie ich do listy
	if (config_file.readBoolEntry("Chat","DefaultTabs"))
	{
		ChatList chList = chat_manager->chats();
		for (uint i = 0; i < chList.count(); i++)
		{
			UserListElements uins = chList[i]->users()->toUserListElements();
			if ((uins.count() > 1 && !config_file.readBoolEntry("Chat","ConferencesInTabs")) || tabdialog->indexOf(chList[i])!=-1)
				continue;
			onNewChat(chList[i]->users());
		}
	}
}

TabsManager::~TabsManager()
{
	kdebugf();

	KaduActions.remove("tab_attach_action");

	UserBox::userboxmenu->removeItem(menuitem);

	ConfigDialog::unregisterSlotOnApplyTab("Chat", this, SLOT(onApplyConfig()));
	ConfigDialog::unregisterSlotOnCreateTab("Chat", this, SLOT(onCreateConfigDialog()));

	ConfigDialog::disconnectSlot("Chat", "Use tabs by default", SIGNAL(toggled(bool)), this, SLOT(onToggledDefaultTabs(bool)));

	ConfigDialog::removeControl("ShortCuts", "Move tab left");
	ConfigDialog::removeControl("ShortCuts", "Move tab right");
	ConfigDialog::removeControl("ShortCuts", "Switch to prev. tab");
	ConfigDialog::removeControl("ShortCuts", "Switch to next tab");
	ConfigDialog::removeControl("ShortCuts", "Tabs");
	ConfigDialog::removeControl("Chat", "Conferences in tabs");
	ConfigDialog::removeControl("Chat", "Auto tab change");
	ConfigDialog::removeControl("Chat", "Tabs below chats");
	ConfigDialog::removeControl("Chat", "Minimum number of tabs");
	ConfigDialog::removeControl("Chat", "Use tabs by default");
	ConfigDialog::removeControl("Chat", "Tabs");

	disconnect(UserBox::userboxmenu, 0, this, 0);
	disconnect(chat_manager, 0, this, 0);
	disconnect(gadu, 0, this, 0);

	saveGeometry(tabdialog, "Chat", "TabWindowsGeometry");

	// jesli kadu nie konczy dzialania to znaczy ze modul zostal tylko wyladowany wiec odlaczamy rozmowy z kart
	if (!Kadu::closing())
	{
		for(int i=tabdialog->count()-1; i>=0; i--)
			detachChat((Chat*)tabdialog->page(i));
	}

	delete tabdialog;
	delete action;
}

void TabsManager::onNewChat(const UserGroup* senders)
{
	kdebugf();
	Chat* chat=chat_manager->findChat(senders);

	if (no_tabs)
	{
		no_tabs=false;
		return;
	}

	bool conferencesInTabs = config_file.readBoolEntry("Chat","ConferencesInTabs");
	if (config_file.readBoolEntry("Chat", "DefaultTabs") && (conferencesInTabs || senders->count() == 1))
	{
		// jesli jest juz otwarte okno z kartami to dodajemy bezwzglednie nowe rozmowy do kart
		if (tabdialog->count() > 0)
			insertTab(chat);
		else if ((newchats.count() + 1) >= config_file.readUnsignedNumEntry("Chat", "MinTabs", 2))
		{
			CONST_FOREACH(ch, newchats)
			{
				// dodajemy karte tylko jesli jej jeszcze nie ma
				if (*ch && tabdialog->indexOf(*ch)==-1)
					insertTab(*ch);
			}
			insertTab(chat);
			newchats.clear();
		}
		else
			newchats.append(chat);
	}
}

void TabsManager::onDestroyingChat(const UserGroup* senders)
{
	kdebugf();
	Chat* chat=chat_manager->findChat(senders);

	if (tabdialog->indexOf(chat) != -1)
		tabdialog->removePage(chat);
	if (tabdialog->count() == 0)
		tabdialog->hide();
	newchats.remove(chat);
	disconnect(chat->edit(), SIGNAL(keyPressed(QKeyEvent*, CustomInput*, bool&)), tabdialog, SLOT(chatKeyPressed(QKeyEvent*, CustomInput*, bool&)));
}

/*
 * XXX: jak bedzie obsluga wielu protokolow to to bedzie trzeba lekko przerobic
 */
void TabsManager::onStatusChanged(UserListElement ule, QString protocolName, const UserStatus &, bool, bool)
{
	kdebugf();
	UserListElements ules;
	ules.append(ule);
	Chat* chat=chat_manager->findChat(ules);

	if (tabdialog->indexOf(chat)!=-1)
	{
		// zdarzenie nie uaktualniala opisu - trzeba o to zadbac
		chat->refreshTitle();

		tabdialog->setTabToolTip(chat, chat->title());
		if (tabdialog->currentPage()==chat)
		{
			tabdialog->setCaption(chat->title());
			tabdialog->setIcon(ule.status(protocolName).pixmap());
		}
		tabdialog->changeTab(chat, ule.status(protocolName).pixmap(), tabdialog->tabLabel(tabdialog->page(tabdialog->indexOf(chat))));
	}
}

void TabsManager::onTabChange(QWidget* w)
{
	Chat* chat = (Chat*)w;

	// czy jest na licie uin-w z nowymi wiadomociami
	if (chatsWithNewMessages.contains(chat))
		chatsWithNewMessages.remove(chat);

	tabdialog->setIcon(*(w->icon()));
	tabdialog->setTabToolTip(chat, chat->title());
	tabdialog->setCaption(chat->title());
	tabdialog->changeTab(chat, *(w->icon()), tabdialog->tabLabel(tabdialog->currentPage()));

}

void TabsManager::onOpenChat(UserListElements senders)
{
	Chat* chat=chat_manager->findChat(senders);

	if (chat && tabdialog->indexOf(chat)!=-1)
	{
		tabdialog->setWindowState(tabdialog->windowState() & ~WindowMinimized);
		tabdialog->setCurrentPage(tabdialog->indexOf(chat));
		tabdialog->raise();
	}
	else if ((config_file.readBoolEntry("Chat","AutoTabChange") && !(chatsWithNewMessages.contains(chat))) ||
		((!tabdialog->isActiveWindow()) && !(chatsWithNewMessages.contains(chat))) ||
		((chatsWithNewMessages.contains(chat)) && !(config_file.readBoolEntry("Chat","OpenChatOnMessage"))))
			autoswith=true;
}

void TabsManager::onChatMsgReceived(Protocol* protocol, UserListElements senders, const QString& msg, time_t time, bool& grab)
{
	kdebugf();
	Chat* chat=chat_manager->findChat(senders);

	if (!(chatsWithNewMessages.contains(chat)) && ((tabdialog->currentPage() != chat) || !tabdialog->isActiveWindow()))
	{
		chatsWithNewMessages.append(chat);
		if (!timer.isActive())
			timer.start(500);
	}
}

void TabsManager::onApplyConfig()
{
	kdebugf();
	if (!config_file.readBoolEntry("Chat", "TabsBelowChats"))
		tabdialog->setTabPosition(QTabWidget::Top);
	else
		tabdialog->setTabPosition(QTabWidget::Bottom);

	// Sprawdzam czy s jakie konferencje a jeli s to ustawiam w nich poprawnie przyciski w zaleznosci
	// czy opcja "Konferencje w kartach" jest wczona/wyczona
	bool conferencesInTabs=config_file.readBoolEntry("Chat", "ConferencesInTabs");
	ChatList chList = chat_manager->chats();
	for (int i = chList.count()-1; i>=0; i--)
	{
		UserListElements uins = chList[i]->users()->toUserListElements();
		if (uins.count() > 1 && tabdialog->indexOf(chList[i])==-1)
		{
			QValueList<ToolButton*> buttons = KaduActions["tab_attach_action"]->toolButtonsForUserListElements(chList[i]->users()->toUserListElements());
			for (QValueList<ToolButton*>::iterator i = buttons.begin(); i != buttons.end(); i++)
				(*i)->setEnabled(conferencesInTabs);
		}
	}
}

void TabsManager::onNewTab()
{
	kdebugf();
	UserBox *activeUserBox = UserBox::activeUserBox();
	if (activeUserBox==NULL)
		return;
	UserListElements users = activeUserBox->selectedUsers();
	bool defaulttabs=config_file.readBoolEntry("Chat", "DefaultTabs");
	Chat* chat=chat_manager->findChat(users);

	// istnieje = przywracamy na pierwszy plan
	if (chat)
	{
		if(tabdialog->indexOf(chat)!=-1)
		{
			tabdialog->setWindowState(tabdialog->windowState() & ~WindowMinimized);
			tabdialog->setCurrentPage(tabdialog->indexOf(chat));
		}
		chat->raise();
		chat->setActiveWindow();
	}
	else
	{
		if (defaulttabs)
			no_tabs=true;

		// nie istnieje to sie stworzy
		chat_manager->openPendingMsgs(users);

		// i rcznie doda
		if (!defaulttabs && (users.size()==1 || config_file.readBoolEntry("Chat","ConferencesInTabs")))
		{
			chat=chat_manager->findChat(users);
			insertTab(chat);
		}
	}
}

void TabsManager::insertTab(Chat* chat)
{
	UserListElements ules = chat->users()->toUserListElements();

	QValueList<ToolButton*> buttons = KaduActions["tab_attach_action"]->toolButtonsForUserListElements(ules);

	for (QValueList<ToolButton*>::iterator i = buttons.begin(); i != buttons.end(); i++)
	{
		(*i)->setOn(true);
		QToolTip::remove(*i);
		QToolTip::add(*i, tr("Detach chat from tabs"));
		if (!(*i)->isEnabled())
			(*i)->setEnabled(true);
	}

	// Ustawiam tytul karty w zaleznosci od tego czy mamy do czynienia z rozmowa czy z konferencja
	if (ules.count()>1)
		tabdialog->addTab(chat, *(chat->icon()), tr("Conference [%1]").arg(ules.count()));
	else
		tabdialog->addTab(chat, *(chat->icon()), ules[0].altNick());

	if ((config_file.readBoolEntry("Chat","AutoTabChange") && !chatsWithNewMessages.contains(chat)) || autoswith)
		tabdialog->setCurrentPage(tabdialog->indexOf(chat));

	tabdialog->setWindowState(tabdialog->windowState() & ~WindowMinimized);
	tabdialog->show();
	tabdialog->raise();
	Chat* ch = (Chat*)tabdialog->currentPage();
	ch->edit()->setFocus();

	if (config_file.readBoolEntry("Chat","ScrollDown"))
		chat->scrollHistoryToBottom();
	autoswith=false;
	connect(chat->edit(), SIGNAL(keyPressed(QKeyEvent*, CustomInput*, bool&)), tabdialog, SLOT(chatKeyPressed(QKeyEvent*, CustomInput*, bool&)));
}

void TabsManager::onPopupMenu()
{
	kdebugf();
	UserBox *activeUserBox = UserBox::activeUserBox();
	if (activeUserBox==NULL)
		return;
	UserListElements users = activeUserBox->selectedUsers();

	if (users.size()!=1 && !config_file.readBoolEntry("Chat", "ConferencesInTabs") && !config_file.readBoolEntry("Chat", "DefaultTabs"))
		UserBox::userboxmenu->setItemEnabled(menuitem, false);

	if (config_file.readBoolEntry("Chat", "DefaultTabs"))
		UserBox::userboxmenu->changeItem(menuitem, tr("Open in new window"));
	else
		UserBox::userboxmenu->changeItem(menuitem, tr("Open in new tab"));

	QString myGGUIN = QString::number(config_file.readNumEntry("General", "UIN"));
	CONST_FOREACH(user, users)
	if ((!(*user).usesProtocol("Gadu")) || (*user).ID("Gadu") == myGGUIN)
		UserBox::userboxmenu->setItemEnabled(menuitem, false);
}

// uff, troche dziwne to ale dziaa tak jak trzeba
void TabsManager::onTimer()
{
	kdebugf();
	Chat* chat;
	static bool msg, wasactive=1;

	// sprawdzaj wszystkie okna ktre s w tabach
	for(int i = tabdialog->count()-1;i>=0;i--)
	{
		chat = (Chat*)tabdialog->page(i);

		// czy trzeba co robi ?
		if (chatsWithNewMessages.contains(chat))
		{
			// okno nieaktywne to trzeba co zrobi
			if (!tabdialog->isActiveWindow())
			{
				if (tabdialog->currentPage() == chat || !msg)
					tabdialog->setCaption(chat->caption()); // bdzie miga tak jak osobne okno
				else if (msg)
					tabdialog->setCaption(tr("NEW MESSAGE(S)"));
			}

			// tab aktualnie nieaktywny to ustaw ikonke
			if (tabdialog->currentPage() != chat)
			{
				if (msg)
					tabdialog->setTabIconSet(chat, icons_manager->loadIcon("Message"));
				else
					tabdialog->setTabIconSet(chat, *(chat->icon()));
			}
			else if (tabdialog->currentPage()==chat && tabdialog->isActiveWindow())
				chatsWithNewMessages.remove(chat); // wywal go z listy uin-w z nowymi wiadomociami

			if (tabdialog->isActiveWindow())
			{
				if (tabdialog->currentPage()==chat)
					tabdialog->setCaption(chat->title()); // a tutaj przywr tytu
				else if (chatsWithNewMessages.count() == 1 && !wasactive && config_file.readBoolEntry("Chat","AutoTabChange"))
					tabdialog->setCurrentPage(tabdialog->indexOf(chat));
			}
		}
	}

	if (chatsWithNewMessages.size()==0)
		timer.stop();

	wasactive = tabdialog->isActiveWindow();
	msg = !msg;
}

void TabsManager::onTabAttach(const UserGroup* users, const QWidget*, bool isOn)
{
	Chat* chat=chat_manager->findChat(users);

	if (!isOn)
		detachChat(chat);
	else
	{
		if (users->count()!=1 && !config_file.readBoolEntry("Chat","ConferencesInTabs"))
			return;
		newchats.clear();
		insertTab(chat);

		QValueList<ToolButton*> buttons = KaduActions["tab_attach_action"]->toolButtonsForUserListElements((chat->users())->toUserListElements());
		for (QValueList<ToolButton*>::iterator i = buttons.begin(); i != buttons.end(); i++)
		{
			QToolTip::remove(*i);
			QToolTip::add(*i, tr("Detach chat from tabs"));
			(*i)->setOn(true);
		}
	}
}

void TabsManager::onContextMenu(QWidget* w, const QPoint& pos)
{
	kdebugf();
	selectedchat=(Chat*)w;
	menu->setItemChecked(4, config_file.readBoolEntry("Chat", "TabsBelowChats"));
	menu->popup(pos);
}

void TabsManager::makePopupMenu()
{
	kdebugf();
	menu=new QPopupMenu();
	menu->setCheckable(true);
	menu->insertItem(icons_manager->loadIcon(dataPath("kadu/modules/data/tabs/attach.png")), tr("Detach"), 0);
	menu->insertItem(tr("Detach all"), 1);
	menu->insertSeparator();
	menu->insertItem(icons_manager->loadIcon("CloseWindow"), tr("Close"), 2);
	menu->insertItem(tr("Close all"), 3);
	menu->insertSeparator();
	menu->insertItem(tr("Tabs below chats"), 4);
	connect(menu, SIGNAL(activated(int)), this, SLOT(onMenu(int)));
}

void TabsManager::onMenu(int id)
{
	switch(id)
	{
		case 0:
			detachChat(selectedchat);
			break;
		case 1:
			for(int i=tabdialog->count()-1; i>=0; i--)
				detachChat((Chat*)tabdialog->page(i));
			break;
		case 2:
			delete selectedchat;
			break;
		case 3:
			for(int i=tabdialog->count()-1; i>=0; i--)
				delete tabdialog->page(i);
			break;
		case 4:
			config_file.writeEntry("Chat", "TabsBelowChats", !menu->isItemChecked(id));
			onApplyConfig();
			break;
	}
}

void TabsManager::onDeleteTab(QWidget* w)
{
	Chat* selectedchat=(Chat*)w;
	delete selectedchat;
}

void TabsManager::onCreateConfigDialog()
{
	ConfigDialog::getSpinBox("Chat", "Minimum number of tabs")
			->setEnabled(ConfigDialog::getCheckBox("Chat", "Use tabs by default")->isChecked());
}

void TabsManager::onToggledDefaultTabs(bool toggled)
{
	ConfigDialog::getSpinBox("Chat", "Minimum number of tabs")->setEnabled(toggled);
}

void TabsManager::onAddedToToolbar(const UserGroup* users)
{
	Chat* chat=chat_manager->findChat(users);
	UserListElements senders=chat->users()->toUserListElements();

		QValueList<ToolButton*> buttons = KaduActions["tab_attach_action"]->toolButtonsForUserListElements((chat->users())->toUserListElements());

	for (QValueList<ToolButton*>::iterator i = buttons.begin(); i != buttons.end(); i++)
	{
		if (senders.size()!=1 && !config_file.readBoolEntry("Chat","ConferencesInTabs") && tabdialog->indexOf(chat)==-1)
		{
			(*i)->setEnabled(false);
			continue;
		}
		if (tabdialog->indexOf(chat)!=-1)
		{
			(*i)->setOn(true);
			QToolTip::remove(*i);
			QToolTip::add(*i, tr("Detach chat from tabs"));
		}
		else
		{
			(*i)->setOn(false);
			QToolTip::remove(*i);
			QToolTip::add(*i, tr("Attach chat to tabs"));
		}
	}
}


bool TabsManager::detachChat(Chat* chat)
{
	if (tabdialog->indexOf(chat)==-1)
		return false;
	UserListElements users=chat->users()->toUserListElements();
	delete chat;
	no_tabs=true;
	chat_manager->openPendingMsgs(users);
	return true;
}

TabsManager* tabs_manager;
